/*
    Blacky HWControl

    Programm zum Zugriff auf IO-Ports

    Copyright (c) Andreas Hofmeier

    This program is free software; you can redistribute it and/or
    modify it under the terms of the GNU General Public License as
    published by the Free Software Foundation; either version 2 of
    the License, or (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
    General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.


    Name:     Andreas Hofmeier

    Adresse:  Stralsunder Strae 20
              28239 Bremen
              Germany

    Telefon:  +49-421-613793

    eMail:    An-H@An-H.de
              hofmeier@inbremen.de
              an-h@gmx.de

    Homepage: http://www.An-H.de/
	      hofmeier.home.pages.de
              http://www.LGut.uni-bremen.de/anh/


    HWCont ist ein kleines Programm, welches den Zugriff auf IO-Ports
    gestattet. Sie knnen als mehrere Ports als Parameter angeben. Das
    Programme berwacht dann diese Ports und gibt alle nderungen aus.
    Sie knnen die Ports natrlich auch beschreiben. Sie knnen mit
    ``=n'' den Port mit der DezimalNummer n beschreiben. Mit ``xn''
    knnen Sie den Port mit der HexaDezimalNummer n beschreiben. Die
    ``~n'' invertiert das Bit des Ports n. Die Bits beginnen bei 0 (0:
    1 1:2 2:4 3:8 4:16 5:32 6:64 7:127). ``+n'' setzt das Bit n und
    ``-n'' lscht das Bit n. Haben Sie mehrere Ports bergeben, knnen
    die Vor das Kommando noch die Nummer des Parameters angeben, so
    beschreiben Sie diesen Port. Wenn Sie also ``hwcont 0x378 0x3bc
    0x278'' aufgerufen haben:

    =20 oder 0=20 oder x14: Set 0x378 to 20
    1+1: Set das Bit 1 auf dem Port 0x3bc
    ...

*/


char blacky_ver_info[] = "Blacky HWControl v0.2 (c) Copyright 2000-2001 Andreas Hofmeier";


#include <unistd.h>
#include <stdio.h>
//#include "/usr/src/linux/include/asm-i386/io.h"
#include <asm/io.h>
#include <time.h>
#include <sys/wait.h>

// Sorry
int i, j;
// Eingabebuffer fr Scanf
unsigned int ui;
// Array von Interern, in dem dir IO-Adresse der Ports stehten
int *ports;
// Array von Bytes in dem die Bytes abgelegt werden, welche vom Port
// gelesen wurden
unsigned char *ports_set;
// Zwischenspeicher fr die Aktuelle Zeit (Die Zeit wird immer mit
// ausgegeben, damit man wei wann sich etwas getan hat)
long ti;
// Zwischenspeicher zum lesen eines Bytes
unsigned char c;
// Globale Kopie von argc, damit printall() wei wieviele Port berwacht werden
int argc_;
// PID des Unterprozesses zum ausgeben der Port-Inhalte
int forkpid = 0;
// Wenn ==1, gibt alle Ports aus, wenn ==0 nur jene, die sich gendert haben
int printall_ignore_last;
// Zykluszeit: 15MiliSecounds wait aufter Scaning
#define WAIT 15

/* Binre wertigkeiten der Bits 1-8 */
char binm[] = {1, 2, 4, 8, 16, 32, 64, 128};

/* Lie das Byte vom IO-Port adr und gib dieses zurck */
unsigned char input(adr)
  int adr;
{
  return inb(adr);
}


/* Schreibe out auf IO-Port adr */
output(adr, out)
  char out;
  int adr;
{
  outb(out, adr);
}


/* Gibt alle Ports aus, welche sich seit dem letztem Zyklus gendert
   haben, wenn printall_ignore_last == 1, alle ausgeben, ohne auf den
   vorherigen Zyklus zu achten */
printall()
{
  for (i = 0; i < (argc_ - 1); i++) {
    c = input(ports[i]);
    if ((c != ports_set[i]) || printall_ignore_last) {
      ports_set[i] = c;
      ti = time(&ti);
      
      printf("< %04d: (0x%04x) ", i, ports[i]);
      
      for (j = 128; j > 0; j = j / 2)
	if ((c & j) > 0) {
	  printf("1");
	} else {
	  printf("0");
	}
	  
      printf(" %03d %02x\t%s", c, c, ctime(&ti));
      fflush(stdout);
    }
  }
}


main(argc, argv)
  int argc;
  char *argv[];
{
  // Argc fr printall() zugnglich machen, das Ding mu wissen
  // wieviele Parameter/Ports das Programm bekam
  argc_ = argc;

  // Version ausgeben
  if ((argc > 1) ? ((argv[1][0] == '-') && (toupper(argv[1][1]) == 'V')) : 0) {
    printf("%s\n\n", blacky_ver_info);
    printkinfo(stdout);
    exit(0);
  }

  // kurtz-Hilfe
  if ((argc < 2) || (argv[1][1] == '?') || (toupper(argv[1][1]) == 'H')) {
      fprintf(stderr, "%s\n\n", blacky_ver_info);
      fprintf(stderr, "Help:\n");
      fprintf(stderr, "  hwcont <port> [port ...]\n");
      fprintf(stderr, "    port  : Schnitstelle. In Dezimal. (hex wird mit 0x eingeleitet)\n");
      fprintf(stderr, "    -v: Version, Licenze\n");

      fprintf(stderr, "--\nGeben Sie ``?'' ein, um Hilfe ber die OnLine-Commandos abzurufen!\n");
      exit(1);
    }

  // 
  if (getuid() != 0) {
    fprintf(stderr, "** ERROR: Dieses Programm mu unter root laufen, da es sonnst nicht auf die HardWare IO-Ports zugreifen kann!\n");
    //    exit(0);
  }

  // Zugriff auf IO-Ports freischalten
  if (iopl(3) < 0) {
    perror("iopl()");
    exit(1);
  }

  // Speicher vom Heab anfordern fr die Liste der Port-Adressen
  ports = (int *) malloc((argc - 1) * sizeof(i));
  if (ports < 1) {
    perror("malloc()");
    exit(1);
  }
  // ... und fr die Port-Inhalte
  ports_set = (unsigned char *) malloc((argc - 1));
  if (ports_set < 1) {
    perror("malloc()");
    exit(1);
  }

  // Laden der Port-Adressen in das Array
  for (i = 0; i < (argc - 1); i++) {
    if ((argv[1 + i][0] == '0') && (argv[1 + i][1] == 'x')) {
      sscanf(argv[1 + i] + 2, "%x", &ports[i]);
    } else {
      sscanf(argv[1 + i], "%d", &ports[i]);
    }
    ports_set[i] = 0;
  }

  // Einmal alle Port-Inhalte ausgeben
  printall_ignore_last = 1;
  printall();
  printall_ignore_last = 0;

  // Scannen der Ports, bei nderung ausgaben
  // ... natrlich im Hintergrund, so das man noch eingaben ttigen
  // kann
  if (!(forkpid = fork())) {
    printf("fork()\n");
    while (1) {
      printall();
      usleep(WAIT);
    }
  }

  // Warte auf Eingabe
  {
    char s[500];
    unsigned char set;
    int port;
    int i, il, ip, ic;

    while(1) {
      // lie eine Zeile von der Eingabe
      gets(s);
      // Wenn leerzeile, EXIT
      if (s[0] == 0) {
	// kille sub-Process
	kill(forkpid, 15);
	while(waitpid(-1,NULL,WNOHANG) > 0);
	// EXIT
	exit(0);
      }
      // Ist eine gltige Eingabe
      il = strlen(s);
      for (i = 0; i < il; i++) {
	if (!isdigit(s[i])) {
	  ic = s[i]; // Welches Commando
	  s[i] = 0;
	  ip = i; // Possition im String
	  // zur Weiterverarbeitung
	  goto OK_ABF;
	}
      }
      // Weiterverarbeitung der Eingabe
    OK_ABF:
      port = 0;
      sscanf(s, "%d", &port);
      set = input(ports[port]);
      sscanf(s + ip + 1, "%d", &i);
      switch(ic) {
	// =: Setzte Dezimal Nummer auf Port
      case '=':
	//	sscanf(s + ip + 1, "%d", &i);
	set = (unsigned char) i;
	break;
	// Einzelnet Bit Setzen (0-7)
      case '+':
	//	sscanf(s + ip + 1, "%d", &i);
	set = set | binm[i];
	break;
	// Einzelnet Bit Lschen (0-7)
      case '-':
	//	sscanf(s + ip + 1, "%d", &i);
	set = set & (~binm[i]);
	break;
	// Einzelnet Bit Invertieren (0-7)
      case '~':
	//	sscanf(s + ip + 1, "%d", &i);
	if ((set & binm[i]) > 0) {
	  set = set & (~binm[i]);
	} else {
	  set = set | binm[i];
	}
	break;
	// x: Setzte HexDezimal Nummer auf Port
      case 'x':
      case 'X':
	sscanf(s + ip + 1, "%x", &ui);
	set = (unsigned char) ui;
	break;
	// Alle Port-Inhalte ausgeben, auch wenn diese sich nicht gendert haben
      case 'p':
      case 'P':
	printall_ignore_last = 1;
	printall();
	printall_ignore_last = 0;
	goto IGNORE;
	break;
	// EXIT
      case 'q':
      case 'Q':
      case 'e':
      case 'E':
	kill(forkpid, 15);
	while(waitpid(-1,NULL,WNOHANG) > 0);
	exit(0);
	goto IGNORE;
	break;
	// HILFE
      case '?':
      case 'H':
	printf("HELP\n");
	printf("  [portnummer]<Command><n>\n");
	printf("  ^           ^\n");
	printf("  |           |\n");
	printf("  |           ? Help\n");
	printf("  |           = Schreibe Dezimalnummer n auf Port\n");
	printf("  |           + Setzte Bit n des Ports (0-7)\n");
	printf("  |           - Lsche Bit n des Ports (0-7)\n");
	printf("  |           ~ Invertiere Bit n des Ports (0-7)\n");
	printf("  |           x Schreibe HexDezimalnummer n auf Port\n");
	printf("  |           p Lie von die Ports und gib das Ergebniss aus\n");
	printf("  |           e Ende\n");
	printf("  Nummer des Parameters, welcher auf den Port zeigt der Bearbeitet werden soll.\n\n");
	printf("  Ausgabe: IO Portnummer HWPortAdresse DualZahl DezimalZahl HexZahl Datum\n");
	printf("           IO: > Ausgabe, schrieb aus IO-Port\n");
	printf("               < Eingabe, la von IO-Port\n");
	fflush(stdout);
	goto IGNORE;
	break;
      default:
	goto IGNORE;
	break;
      } /* switch() */
      // OutPut: Schreibe den neuen Inhalt auf den Port
      output(ports[port], set);
      // Gib die Aktion auf dem Bildschirm aus
      ti = time(&ti);
      printf("> %04d: (0x%04x) ", port, ports[port]);
      for (i = 128; i > 0; i = i / 2)
	if ((set & i) > 0) {
	  printf("1");
	} else {
	  printf("0");
	}
      printf(" %03d %02x\t%s", set, set, ctime(&ti));
      fflush(stdout);
    IGNORE:
    }
  }
}  
